@@ -25,28 +25,66 @@ unsigned debug_level;
 int debug_option;
 struct rush_rule *rule_head, *rule_tail;
 struct passwd *rush_pw;
+char *rush_interactive_shell;
 
 #define STDOUT_FILENO 1
 #define STDERR_FILENO 2
 
-char *error_msg[] = {
-        /* usage_error */
-        N_("You are not permitted to execute this command.\n"             
-           "Contact the systems administrator for further assistance.\n"),
-        
-        /* nologin_error */
-        N_("You do not have interactive login access to this machine.\n" 
-           "Contact the systems administrator for further assistance.\n"),
-        
-        /* config_error */
-        N_("Local configuration error occurred.\n" 
-           "Contact the systems administrator for further assistance.\n"),
-        
-        /* system_error */
-        N_("A system error occurred while attempting to execute command.\n" 
-           "Contact the systems administrator for further assistance.\n")
+struct error_msg {
+	char *text;          /* Message text */
+	int custom;          /* True, if the message was customized */
 };
 
+struct error_msg error_msg[] = {
+        /* usage_error */
+	{ N_("You are not permitted to execute this command.\n"             
+	     "Contact the systems administrator for further assistance.\n"), },
+        
+        /* nologin_error */
+        { N_("You do not have interactive login access to this machine.\n" 
+	     "Contact the systems administrator for further assistance.\n") },
+        
+        /* config_error */
+        { N_("Local configuration error occurred.\n" 
+	     "Contact the systems administrator for further assistance.\n") },
+        
+        /* system_error */
+        { N_("A system error occurred while attempting to execute command.\n" 
+	     "Contact the systems administrator for further assistance.\n") }
+};
+
+void
+set_error_msg(enum error_type type, char *text)
+{
+	error_msg[type].text = text;
+	error_msg[type].custom = 1;
+}
+
+int
+string_to_error_index(const char *name)
+{
+	static const char *error_msg_name[] = {
+		"usage-error",
+		"nologin-error",
+		"config-error",
+		"system-error",
+		NULL
+	};
+	static int error_msg_index[] = {
+		usage_error,
+		nologin_error,
+		config_error,
+		system_error
+	};
+	ARGMATCH_VERIFY(error_msg_name, error_msg_index);
+
+	ptrdiff_t d = ARGMATCH(name, error_msg_name, error_msg_index);
+	if (d < 0)
+		return -1;
+	return error_msg_index[d];
+}
+	
+
 void
 send_msg(const char *msg, size_t len)
 {
@@ -102,14 +140,17 @@ die(enum error_type type, struct rush_i18n *i18n, const char *fmt, ...)
         vlogmsg(LOG_ERR, fmt, ap);
         va_end(ap);
 	if (!lint_option) {
-		const char *msg;
-		if (i18n) 
-			msg = user_gettext(i18n->locale,
-					   i18n->text_domain,
-					   i18n->localedir,
-					   error_msg[type]);
-		else
-			msg = error_msg[type];
+		const char *msg = error_msg[type].text;
+		if (error_msg[type].custom) {
+			/* If it is a customized version, translate it via
+			   user-supplied i18n */
+			if (i18n) 
+				msg = user_gettext(i18n->locale,
+						   i18n->text_domain,
+						   i18n->localedir,
+						   msg);
+		} else
+			msg = gettext(msg);
 		send_msg(msg, strlen(msg));
 		sleep(sleep_time);
 	}
@@ -159,13 +200,13 @@ test_request_cmdline(struct test_node *node, struct rush_request *req)
         return regexec(&node->v.regex, req->cmdline, 0, NULL, 0);
 }
 
+#define ARG_NO(n,argc) (((n) < 0) ? (argc) + (n) : (n))
+
 int
 test_request_arg(struct test_node *node, struct rush_request *req)
 {
-        int n = node->v.arg.arg_no;
-        if (n == -1)
-                n = req->argc - 1;
-        if (n >= req->argc)
+        int n = ARG_NO(node->v.arg.arg_no, req->argc);
+        if (n < 0 || n >= req->argc)
                 return 1;
         return regexec(&node->v.arg.regex, req->argv[n], 0, NULL, 0);
 }
@@ -466,6 +507,8 @@ reparse_cmdline(struct rush_request *req)
         if ((rc = argcv_get(req->cmdline, NULL, NULL, &req->argc, &req->argv)))
                 die(system_error, &req->i18n, _("argcv_get(%s) failed: %s"),
                     req->cmdline, strerror(rc));
+	free(req->prog);
+	req->prog = NULL;
 }
 
 void
@@ -479,14 +522,53 @@ rebuild_cmdline(struct rush_request *req)
 		    _("argcv_string failed: %s"), strerror(rc));
 }
 
+static int
+get_arg_no(int index, struct rush_request *req)
+{
+	int arg_no = ARG_NO(index, req->argc);
+	if (arg_no < 0 || arg_no >= req->argc) 
+		die(config_error,
+		    &req->i18n, 
+		    _("no argument at index %d in command: %s"),
+		    index,
+		    req->cmdline);
+	return arg_no;
+}
+
+static void
+assign_string(char **pstr, char *val)
+{
+	debug2(2, _("Transform: \"%s\" -> \"%s\""), *pstr ? *pstr : "", val);
+	free(*pstr);
+	*pstr = val;
+}
+
 void
 run_transforms(struct rush_rule *rule, struct rush_request *req)
 {
         struct transform_node *node;
         char *p;
-        int arg_no;
+        int i, arg_no, arg_end;
         int args_transformed = 0;
-        
+	char *val, **target;
+	
+#define GET_TGT_VAL()							\
+	if (node->progmod) {						\
+		arg_no = 0;						\
+		target = &req->prog;					\
+		val = PROGFILE(req);					\
+		debug1(2, _("Modifying program name (%s)"), val);	\
+	} else {							\
+		arg_no = get_arg_no(node->arg_no, req);			\
+		target = &req->argv[arg_no];				\
+		val = *target;						\
+		args_transformed = 1;					\
+		debug1(2, _("Modifying argv[%d]"), arg_no);		\
+	} 								\
+	if (node->pattern) {						\
+		val = rush_expand_string(node->pattern, req);		\
+	}
+
         for (node = rule->transform_head; node; node = node->next) {
                 switch (node->type) {
                 case transform_cmdline:
@@ -495,26 +577,76 @@ run_transforms(struct rush_rule *rule, struct rush_request *req)
                                 args_transformed = 0;
                         }
                         debug(2, _("Transforming command line"));
-                        p = transform_string(node->trans, req->cmdline);
+                        p = transform_string(node->v.trans, req->cmdline);
                         free(req->cmdline);
                         req->cmdline = p;
                         debug1(2, _("Command line: %s"), req->cmdline);
                         reparse_cmdline(req);
                         break;
 
+		case transform_setcmd:
+			if (args_transformed) {
+                                rebuild_cmdline(req);
+                                args_transformed = 0;
+                        }
+                        debug(2, _("Setting command line"));
+                        free(req->cmdline);
+                        req->cmdline = xstrdup(val);
+                        debug1(2, _("Command line: %s"), req->cmdline);
+                        reparse_cmdline(req);
+                        break;
+			
                 case transform_arg:
-                        arg_no = node->arg_no;
-                        if (arg_no == -1)
-                                arg_no = req->argc - 1;
-                        if (arg_no >= req->argc) 
-                                die(usage_error,
-				    &req->i18n, 
-                                    _("not enough arguments in command: %s"),
-                                    req->cmdline);
-                        p = transform_string(node->trans, req->argv[arg_no]);
-                        free(req->argv[arg_no]);
-                        req->argv[arg_no] = p;
-                        args_transformed = 1;
+			GET_TGT_VAL();
+                        p = transform_string(node->v.trans, val);
+			assign_string(target, p);
+			break;
+
+		case transform_map:
+			GET_TGT_VAL();
+			debug6(2,
+			       _("Transformation map: %s, %s, %s, %u, %u, %s"),
+			       node->v.map.file,
+			       node->v.map.delim,
+			       node->v.map.key,
+			       node->v.map.key_field,
+			       node->v.map.val_field,
+			       node->v.map.defval);
+                        p = map_string(&node->v.map, req);
+			if (p) 
+				assign_string(target, p);
+			else
+				args_transformed = 0;
+			break;
+
+		case transform_delarg:
+			arg_no = get_arg_no(node->arg_no, req);
+			arg_end = get_arg_no(node->v.arg_end, req);
+			if (arg_end < arg_no) {
+				int x = arg_end;
+				arg_end = arg_no;
+				arg_no = x;
+			}
+			debug2(2, _("Deleting arguments %d-%d"),
+			       arg_no, arg_end);
+			if (arg_no == 0 || arg_end == 0)
+				die(config_error,
+				    &req->i18n, _("Deleting argv[0] is prohibited"));
+			for (i = arg_no; i <= arg_end; i++) 
+				free(req->argv[i]);
+			i = arg_end - arg_no + 1;
+			memmove(req->argv + arg_no,
+				req->argv + arg_end + 1,
+				(req->argc - i)
+				  * sizeof(req->argv[0]));
+			req->argc -= i;
+			args_transformed = 1;
+			break;
+
+		case transform_setarg:
+			GET_TGT_VAL();
+			assign_string(target, val);
+			break;
                 }
         }
 
@@ -523,6 +655,7 @@ run_transforms(struct rush_rule *rule, struct rush_request *req)
 
         if (__debug_p(2)) {
                 int i;
+		logmsg(LOG_DEBUG, _("Program name: %s"), PROGFILE(req));
                 logmsg(LOG_DEBUG, _("Final arguments:"));
                 for (i = 0; i < req->argc; i++)
                         logmsg(LOG_DEBUG, "% 4d: %s", i, req->argv[i]);
@@ -624,16 +757,35 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 		req->i18n.locale = rule->i18n.locale;
 
         if (rule->error_msg) {
-		const char *msg;
+		const char *msg = rule->error_msg;
+		int custom = 1;
+		
+                debug1(2, _("Error message: %s"), msg);
+		if (msg[0] == '@') {
+			int n;
+			
+			if (msg[1] == '@')
+				msg++;
+			else if ((n = string_to_error_index(msg + 1)) == -1) 
+				logmsg(LOG_NOTICE,
+				       _("Unknown message reference: %s\n"),
+				       msg);
+			else {
+				msg = error_msg[n].text;
+				custom = error_msg[n].custom;
+			}
+		} 
 
-                debug1(2, _("Error message: %s"), rule->error_msg);
-		msg = user_gettext(rule->i18n.locale,
-				   rule->i18n.text_domain,
-				   rule->i18n.localedir,
-				   rule->error_msg);
+		if (custom) 
+			msg = user_gettext(rule->i18n.locale,
+					   rule->i18n.text_domain,
+					   rule->i18n.localedir,
+					   msg);
+		else
+			msg = gettext(msg);
                 if (write(rule->error_fd, msg, strlen(msg)) < 0)
                         die(system_error, &req->i18n, 
-                            _("Error sending error message to descriptor %d: %s"),
+                            _("Error sending diagnostic message to descriptor %d: %s"),
                             rule->error_fd, strerror(errno));
                 exit(1);
         }
@@ -700,7 +852,8 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
                 die(system_error, &req->i18n, _("cannot change to dir %s: %s"),
                     req->home_dir, strerror(errno));
 
-        debug1(2, _("Executing %s"), req->cmdline);
+        debug2(2, _("Executing %s, %s"),
+	       PROGFILE(req), req->cmdline);
 	if (lint_option)
 		exit(0);
 
@@ -709,7 +862,7 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 
         if (setuid(req->pw->pw_uid))
                 die(system_error, &req->i18n, _("cannot enforce uid %lu: %s"),
-                    req->pw->pw_uid, strerror(errno));
+                    (unsigned long) req->pw->pw_uid, strerror(errno));
 
 	if (req->pw->pw_uid && setuid(0) == 0) 
 		die(system_error, &req->i18n,
@@ -717,7 +870,7 @@ run_rule(struct rush_rule *rule, struct rush_request *req)
 	
         umask(req->umask);
 
-        execve(req->argv[0], req->argv, new_env);
+        execve(PROGFILE(req), req->argv, new_env);
         die(system_error, &req->i18n, _("%s:%d: %s: cannot execute %s: %s"),
             rule->file, rule->line, rule->tag,
             req->cmdline, strerror(errno));
@@ -773,7 +926,12 @@ main(int argc, char **argv)
 	if (!command) {
 		if (lint_option) 
 			exit(0);
-		die(usage_error, NULL, _("invalid command line"));
+		else if (rush_interactive_shell) {
+			command = rush_interactive_shell;
+			debug1(2, _("Mapping interactive shell to \"%s\""),
+			       command);
+		} else 
+			die(usage_error, NULL, _("invalid command line"));
 	}
 
         if (__debug_p(2)) {
