@@ -33,6 +33,7 @@
 #include <fcntl.h>
 #include <signal.h>
 #include <getopt.h>
+#include <ctype.h>
 
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -45,6 +46,7 @@
 #include <c-ctype.h>
 #include <argcv.h>
 #include <inttostr.h>
+#include "argmatch.h"
 
 #include <defines.h>
 #include <librush.h>
@@ -79,16 +81,35 @@ enum error_type {
 typedef struct limits_rec *limits_record_t;
 typedef struct transform *transform_t;
 
+struct rush_map {
+	char *file;
+	char *delim;
+	char *key;
+	unsigned key_field;
+	unsigned val_field;
+	char *defval;
+};
+	
 enum transform_node_type {
 	transform_cmdline,
-	transform_arg
+	transform_arg,
+	transform_map,
+	transform_delarg,
+	transform_setcmd,
+	transform_setarg
 };
 
 struct transform_node {
 	struct transform_node *next;
 	enum transform_node_type type;
-	transform_t trans;
 	int arg_no;
+	int progmod;
+	char *pattern;
+	union {
+		transform_t trans;
+		struct rush_map map;
+		int arg_end;
+	} v;
 };
 
 /* Comparison operator */
@@ -183,12 +204,14 @@ struct rush_rule {
 };
 
 struct rush_request {
-        char *cmdline;
-        int argc;
-        char **argv;
-        struct passwd *pw;
-	unsigned umask;
-	char *chroot_dir;
+        char *cmdline;         /* Command line */
+        int argc;              /* Number of elements in argv */
+        char **argv;           /* Command line in parsed form */
+	char *prog;            /* Program file name, if different
+				  from argv[0] */
+        struct passwd *pw;     /* User passwd entry */
+	unsigned umask;        
+	char *chroot_dir;      
         char *home_dir;
 	enum rush_three_state fork;
 	enum rush_three_state acct;
@@ -196,16 +219,17 @@ struct rush_request {
 	struct rush_i18n i18n;
 };
 
+#define PROGFILE(req) ((req)->prog ? (req)->prog : (req)->argv[0])
 #define NO_UMASK ((mode_t)-1)
 
 extern char *rush_config_file;
 extern int lint_option;
 extern unsigned sleep_time;
-extern char *error_msg[];
 extern struct rush_rule *rule_head, *rule_tail;
 extern unsigned debug_level;
 extern int debug_option;
 extern struct passwd *rush_pw;
+extern char *rush_interactive_shell;
 
 #define __debug_p(x) ((x) <= debug_level)
 
@@ -242,12 +266,18 @@ extern struct passwd *rush_pw;
 			logmsg(LOG_DEBUG, fmt, x1, x2,	\
 			       x3, x4, x5);		\
 	} while(0)
-#define debug6(lev,fmt,x1,x2,x3,x4,x5)		        \
+#define debug6(lev,fmt,x1,x2,x3,x4,x5,x6)	        \
 	do {						\
 		if (__debug_p(lev))			\
 			logmsg(LOG_DEBUG, fmt, x1, x2,	\
 			       x3, x4, x5, x6);		\
 	} while(0)
+#define debug7(lev,fmt,x1,x2,x3,x4,x5,x6,x7)	        \
+	do {						\
+		if (__debug_p(lev))			\
+			logmsg(LOG_DEBUG, fmt, x1, x2,	\
+			       x3, x4, x5, x6, x7);	\
+	} while(0)
 
 void die(enum error_type type, struct rush_i18n *i18n, const char *fmt, ...)
 	 ATTRIBUTE_NORETURN RUSH_PRINTFLIKE(3,4);
@@ -258,6 +288,9 @@ int set_user_limits(const char *name, struct limits_rec *lrec);
 
 void parse_config(void);
 
+void set_error_msg(enum error_type type, char *text);
+int string_to_error_index(const char *name);
+
 transform_t compile_transform_expr (const char *expr);
 char *transform_string (struct transform *tf, const char *input);
 
@@ -284,7 +317,11 @@ char *expand_tilde(const char *dir, const char *home);
 #endif
 
 int check_config_permissions(const char *filename, struct stat *st);
-int cfck_keyword(const char *name, size_t len);
+int cfck_keyword(const char *name);
 
+
+/* map.c */
+char *map_string(struct rush_map *map, struct rush_request *req);
+char *rush_expand_string(const char *string, struct rush_request *req);
 
 
